/*******************************************************************************
*    IM-Daemon                                                                 *
*    Copyright (C) 2012-averyfarwaydate Luca Bertoncello                       *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#define _GNU_SOURCE
#include <string.h>
#include <glib.h>
#include <unistd.h>
#include <gd.h>
#include "incs/config.h"
#include "incs/logger.h"
#include "incs/im.h"

extern int purpleDebug;
PurpleAccount *account;
gboolean signedOn, connecting;
char iconPath[255];

static PurpleCoreUiOps coreUIops =
{
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL
};

static PurpleConnectionUiOps connectionUIops =
{
  NULL,                      /* connect_progress         */
  NULL,                      /* connected                */
  NULL,                      /* disconnected             */
  NULL,                      /* notice                   */
  NULL,                      /* report_disconnect        */
  NULL,                      /* network_connected        */
  networkDisconnectedCB,     /* network_disconnected     */
  reportDisconnectReasonCB,  /* report_disconnect_reason */
  NULL,
  NULL,
  NULL
};

static PurpleEventLoopUiOps glibEventloops =
{
  g_timeout_add,
  g_source_remove,
  glibInputAddCB,
  g_source_remove,
  NULL,
#if GLIB_CHECK_VERSION(2,14,0)
  g_timeout_add_seconds,
#else
  NULL,
#endif
  NULL,
  NULL,
  NULL
};

/**
 * Nötige Funktion für die Verwaltung der Meldungen vom Server
 */
void glibIODestroy(gpointer data)
{
  g_free(data);
}

/**
 * Nötige Funktion für die Verwaltung der Meldungen vom Server
 */
gboolean glibIOInvoke(GIOChannel *source, GIOCondition condition, gpointer data)
{
  PurpleGLibIOClosure *closure = data;
  PurpleInputCondition purpleCond = 0;

  if (condition & PURPLE_GLIB_READ_COND)
    purpleCond |= PURPLE_INPUT_READ;
  if (condition & PURPLE_GLIB_WRITE_COND)
    purpleCond |= PURPLE_INPUT_WRITE;

  closure->function(closure->data, g_io_channel_unix_get_fd(source), purpleCond);

  return TRUE;
}

/**
 * Nötige Funktion für die Verwaltung der Meldungen vom Server
 */
guint glibInputAddCB(gint fd, PurpleInputCondition condition, PurpleInputFunction function, gpointer data)
{
  PurpleGLibIOClosure *closure = g_new0(PurpleGLibIOClosure, 1);
  GIOChannel *channel;
  GIOCondition cond = 0;

  closure->function = function;
  closure->data = data;

  if (condition & PURPLE_INPUT_READ)
    cond |= PURPLE_GLIB_READ_COND;
  if (condition & PURPLE_INPUT_WRITE)
    cond |= PURPLE_GLIB_WRITE_COND;

  channel = g_io_channel_unix_new(fd);
  closure->result = g_io_add_watch_full(channel, G_PRIORITY_DEFAULT, cond, glibIOInvoke, closure, glibIODestroy);

  g_io_channel_unref(channel);
  return closure->result;
}

/**
 * Initialisiert die LibPurple
 */
void initPurple()
{
  int handle;

  signedOn = connecting = FALSE;

  LOGGER(LOGL_DEBUG, "Initialising Libpurple");
  purple_util_set_user_dir(CUSTOM_USER_DIRECTORY);
  purple_debug_set_enabled(FALSE);
  purple_core_set_ui_ops(&coreUIops);
  purple_eventloop_set_ui_ops(&glibEventloops);
  purple_connections_set_ui_ops(&connectionUIops);
  purple_plugins_add_search_path(CUSTOM_PLUGIN_PATH);

  if(!purple_core_init(UI_ID))
    LOGGER(LOGL_FATAL, "Initialisation Libpurple failed");

  purple_set_blist(purple_blist_new());
  purple_blist_load();
  purple_prefs_load();
  purple_plugins_load_saved(PLUGIN_SAVE_PREF);
  purple_pounces_load();
  LOGGER(LOGL_INFO, "Libpurple initialised");

// http://libpurple.com/category/getting-your-hands-dirty/
  purple_signal_connect(purple_connections_get_handle(), "signed-on", &handle,
      PURPLE_CALLBACK(signedOnCB), NULL);
  purple_signal_connect(purple_accounts_get_handle(), "account-authorization-requested", &handle,
          PURPLE_CALLBACK(authRequestedCB), NULL);
}

/**
 * Setzt die eigene Icon
 *
 * @param PurpleAccount *account          Pidgin-Konto
 * @param char *icon                      Pfad zur Icon
 */
void setMyIcon(PurpleAccount *account, char *icon)
{
  size_t len;
  gpointer data;
  PurpleStoredImage *img;
  FILE *fp;
  gdImagePtr im = NULL;
  char tmpFile[255];
  int fd;

  LOGGER(LOGL_INFO, "Setting buddy icon [%s]", icon);

  fp = fopen(icon, "rb");
  if(strcasestr(icon, ".jpg") != NULL)
    im = gdImageCreateFromJpeg(fp);
  else if(strcasestr(icon, ".gif") != NULL)
    im = gdImageCreateFromGif(fp);
  else if(strcasestr(icon, ".png") != NULL)
    im = gdImageCreateFromPng(fp);
  else if(strcasestr(icon, ".xbm") != NULL)
    im = gdImageCreateFromXbm(fp);
  else if(strcasestr(icon, ".xpm") != NULL)
    im = gdImageCreateFromXpm(icon);
  fclose(fp);

  strcpy(tmpFile, "/tmp/icon-XXXXXX");
  fd = mkstemp(tmpFile);
  close(fd);
  fp = fopen(tmpFile, "wb");
  if(strcmp(account->protocol_id, "prpl-aim") == 0)
  {
    LOGGER(LOGL_DEBUG, "Converting icon to JPG");
    gdImageJpeg(im, fp, 95);
  }
  else if(strcmp(account->protocol_id, "prpl-icq") == 0)
  {
    LOGGER(LOGL_DEBUG, "Converting icon to JPG");
    gdImageJpeg(im, fp, 95);
  }
  else if(strcmp(account->protocol_id, "prpl-msn") == 0)
  {
    LOGGER(LOGL_DEBUG, "Converting icon to PNG");
    gdImagePng(im, fp);
  }
  else if(strcmp(account->protocol_id, "prpl-jabber") == 0)
  {
    LOGGER(LOGL_DEBUG, "Converting icon to PNG");
    gdImagePng(im, fp);
  }
  else if(strcmp(account->protocol_id, "prpl-yahoo") == 0)
  {
    LOGGER(LOGL_DEBUG, "Converting icon to JPG");
    gdImageJpeg(im, fp, 95);
  }
  else
  {
    LOGGER(LOGL_DEBUG, "Converting icon to JPG");
    gdImageJpeg(im, fp, 95);
  }
  fclose(fp);

  LOGGER(LOGL_DEBUG, "Using %s", tmpFile);
  img = purple_imgstore_new_from_file(tmpFile);
  len = purple_imgstore_get_size(img);
  data = g_memdup(purple_imgstore_get_data(img), len);
  purple_account_set_buddy_icon_path(account, tmpFile);
  purple_buddy_icons_set_account_icon(account, data, len);
  unlink(tmpFile);
}

/**
 * CallBack für die Anfrage nach Autorisierung
 */
int authRequestedCB(PurpleAccount *account, const char *user)
{
  LOGGER(LOGL_INFO, "Buddy '%s' request authorisation", user);
  return 1; // immer autorisieren. Wenn -1 zurückgegeben, immer Autorisierung abgelehnt
}

/**
 * CallBack nach dem erfolgreichen Login
 */
void signedOnCB(PurpleConnection *gc, gpointer null)
{
  PurpleAccount *account = purple_connection_get_account(gc);

  LOGGER(LOGL_INFO, "Account connected: %s", account->username);
  signedOn = TRUE;
  connecting = FALSE;
  if(strlen(iconPath) != 0)
    setMyIcon(account, iconPath);
}

/**
 * CallBack nach einer Trennung des Netzes
 */
void networkDisconnectedCB()
{
  LOGGER(LOGL_INFO, "Disconnected from remote server");
}

/**
 * CallBack nach einer Trennung des Netzes
 */
void reportDisconnectReasonCB(PurpleConnection *gc, PurpleConnectionError reason, const char *reasonText)
{
  signedOn = FALSE;
// Wenn es nicht geschafft hat, sich zu verbinden, muss "connecting" wieder auf FALSE gesetzt werden
  if(connecting)
    connecting = FALSE;
  LOGGER(LOGL_INFO, "Connection disconnected. Reason: %s", reasonText);
}

/**
 * Verbindet mit dem Server
 *
 * @param struct config cfg     Die Konfiguration
 */
void connectToIMServer(struct config cfg)
{
  PurpleSavedStatus *status;
  char proto[255];
  gboolean validIcon = FALSE;

  if(purpleDebug == 1)
    purple_debug_set_enabled(TRUE);
  connecting = TRUE;
  LOGGER(LOGL_INFO, "Connecting to remote server");
  bzero(proto, 255);
  sprintf(proto, "prpl-%s", cfg.protocol);

  bzero(iconPath, 255);
  account = purple_account_new(cfg.login, proto);
  purple_account_set_password(account, cfg.password);
  purple_account_set_string(account, "server", cfg.server);
  purple_account_set_int(account, "port", cfg.port);
  if(cfg.iconPath != NULL)
  {
// https://developer.pidgin.im/wiki/Using%20Pidgin#HowdoIsetabuddyiconoravatar
    if(strcasestr(cfg.iconPath, ".gif") != NULL)
      validIcon = TRUE;
    if(strcasestr(cfg.iconPath, ".jpg") != NULL)
      validIcon = TRUE;
    if(strcasestr(cfg.iconPath, ".xbm") != NULL)
      validIcon = TRUE;
    if(strcasestr(cfg.iconPath, ".xpm") != NULL)
      validIcon = TRUE;
    if(strcasestr(cfg.iconPath, ".png") != NULL)
      validIcon = TRUE;
    if(validIcon == FALSE)
      LOGGER(LOGL_NOTICE, "Icon is NOT a known format. Only GIF, JPG, XBM, XPM and PNG are supported. Ignoring icon...");
    else
    {
      strcpy(iconPath, cfg.iconPath);
      setMyIcon(account, iconPath);
    }
  }
  if(cfg.requireTLS)
    purple_account_set_bool(account, "require_tls", TRUE);
  else
    purple_account_set_bool(account, "require_tls", FALSE);

  purple_account_set_bool(account, "auth_plain_in_clear", TRUE);
  purple_account_set_enabled(account, UI_ID, TRUE);
  status = purple_savedstatus_new("ONLINE", PURPLE_STATUS_AVAILABLE);
  purple_savedstatus_activate(status);
}

/**
 * Trennt die Verbindung vom Server
 */
void disconnectFromIMServer()
{
  LOGGER(LOGL_INFO, "Disconnecting from remote server");
  purple_account_disconnect(account);
}

/**
 * Schicke eine IM-Meldung einem Empfängern
 *
 * @param char *recipient       Der Empfänger
 * @param char *message         Die Meldung
 */
void sendIM(char *recipient, char *message)
{
  PurpleConversation *pc;

  if(signedOn)
  {
    LOGGER(LOGL_DEBUG, "Sending message to %s: [%s]", recipient, message);
    pc = purple_conversation_new(PURPLE_CONV_TYPE_IM, account, recipient);
    purple_conv_im_send(PURPLE_CONV_IM(pc), message);
  }
  else
    LOGGER(LOGL_NOTICE, "Message to %s NOT sent, because not yet connected", recipient);
}

/**
 * Gibt zurück, ob die Verbindung mit dem Server besteht
 *
 * @return gboolean             TRUE wenn die Verbindung besteht,
 *                              FALSE wenn nicht
 */
gboolean isSignedOn()
{
  return signedOn;
}

/**
 * Gibt zurück, ob gerade versucht wird, sich mit dem Server zu verbinden
 *
 * @return gboolean             TRUE wenn eine Verbingung aufgebaut wird,
 *                              FALSE wenn nicht
 */
gboolean isConnecting()
{
  return connecting;
}
